# Copyright 2015 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from itertools import groupby
import os
import re
import sys
import time
from xml.sax.saxutils import escape
from xml.sax.saxutils import quoteattr

from ament_copyright import CONTRIBUTING_FILETYPE
from ament_copyright import get_copyright_names
from ament_copyright import get_licenses
from ament_copyright import LICENSE_FILETYPE
from ament_copyright import SOURCE_FILETYPE
from ament_copyright import UNKNOWN_IDENTIFIER
from ament_copyright.crawler import get_files
from ament_copyright.parser import get_comment_block
from ament_copyright.parser import get_index_of_next_line
from ament_copyright.parser import parse_file
from ament_copyright.parser import scan_past_coding_and_shebang_lines
from ament_copyright.parser import scan_past_empty_lines
from ament_copyright.parser import search_copyright_information


def main(argv=sys.argv[1:]):
    extensions = [
        'c', 'cc', 'cpp', 'cxx', 'h', 'hh', 'hpp', 'hxx',
        'cmake',
        'py',
    ]

    parser = argparse.ArgumentParser(
        description='Check code files for copyright and license information.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'paths',
        nargs='*',
        default=[os.curdir],
        help='The files or directories to check. For directories files ending '
             "in %s will be considered (except directories starting with '.' "
             "or '_' and 'setup.py' files beside 'package.xml' files)." %
             ', '.join(["'.%s'" % e for e in extensions]))
    parser.add_argument(
        '--exclude',
        metavar='filename',
        nargs='*',
        default=[],
        dest='excludes',
        help='The filenames to exclude.')
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--add-missing',
        nargs=2,
        metavar=('COPYRIGHT_NAME', 'LICENSE'),
        help=(
            'Add missing copyright notice and license information using the '
            'passed copyright holder and license'))
    group.add_argument(
        '--add-copyright-year',
        nargs='*',
        type=int,
        help='Add the current year to existing copyright notices')
    group.add_argument(
        '--list-copyright-names',
        action='store_true',
        help='List names of known copyright holders')
    group.add_argument(
        '--list-licenses',
        action='store_true',
        help='List names of known licenses')
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='Show all files instead of only the ones with errors / modifications')
    # not using a file handle directly
    # in order to prevent leaving an empty file when something fails early
    group.add_argument(
        '--xunit-file',
        help='Generate a xunit compliant XML file')
    args = parser.parse_args(argv)

    names = get_copyright_names()
    if args.list_copyright_names:
        for key in sorted(names.keys()):
            print('%s: %s' % (key, names[key]))
        return 0

    licenses = get_licenses()
    if args.list_licenses:
        for key in sorted(licenses.keys()):
            print('%s: %s' % (key, licenses[key].name))
        return 0

    if args.xunit_file:
        start_time = time.time()

    filenames = get_files(args.paths, extensions, args.excludes)
    if not filenames:
        print('No repository roots and files found')

    file_descriptors = {}
    for filename in sorted(filenames):
        file_descriptors[filename] = parse_file(filename)

    if args.add_missing:
        name = names.get(args.add_missing[0], args.add_missing[0])
        if args.add_missing[1] not in licenses:
            parser.error(
                "'LICENSE' argument must be a known license name. "
                "Use the '--list-licenses' options to see alist of valid license names.")
        add_missing_header(
            file_descriptors, name, licenses[args.add_missing[1]], args.verbose)
        return 0

    if args.add_copyright_year is not None:
        if not args.add_copyright_year:
            args.add_copyright_year.append(time.strftime('%Y'))
        args.add_copyright_year = [int(year) for year in args.add_copyright_year]
        add_copyright_year(file_descriptors, args.add_copyright_year, args.verbose)
        return 0

    report = []

    # check each directory for CONTRIBUTING.md and LICENSE files
    for path in sorted(file_descriptors.keys()):
        file_descriptor = file_descriptors[path]
        message = None
        has_error = False

        if file_descriptor.filetype == SOURCE_FILETYPE:
            if not file_descriptor.exists:
                message = 'file not found'
                has_error = True

            elif not file_descriptor.content:
                message = 'file empty'

            elif not file_descriptor.copyright_identifiers:
                message = 'could not find copyright notice'
                has_error = True

            else:
                message = 'copyright=%s, license=%s' % \
                    (', '.join([str(c) for c in file_descriptor.copyrights]),
                     file_descriptor.license_identifier)
                has_error = file_descriptor.license_identifier == UNKNOWN_IDENTIFIER

        elif file_descriptor.filetype in [CONTRIBUTING_FILETYPE, LICENSE_FILETYPE]:
            if not file_descriptor.exists:
                message = 'file not found'
                has_error = True

            elif not file_descriptor.content:
                message = 'file empty'
                has_error = True

            elif file_descriptor.license_identifier:
                message = file_descriptor.license_identifier
                has_error = file_descriptor.license_identifier == UNKNOWN_IDENTIFIER

            else:
                assert False, file_descriptor

        else:
            assert False, 'Unknown filetype: ' + file_descriptor.filetype

        if args.verbose or has_error:
            print('%s: %s' % (file_descriptor.path, message),
                  file=sys.stderr if has_error else sys.stdout)
        report.append((file_descriptor.path, not has_error, message))

    # output summary
    error_count = len([r for r in report if not r[1]])
    if not error_count:
        print('No problems found, checked %d files' % len(report))
        rc = 0
    else:
        print('%d errors, checked %d files' % (error_count, len(report)), file=sys.stderr)
        rc = 1

    # generate xunit file
    if args.xunit_file:
        folder_name = os.path.basename(os.path.dirname(args.xunit_file))
        file_name = os.path.basename(args.xunit_file)
        suffix = '.xml'
        if file_name.endswith(suffix):
            file_name = file_name[0:-len(suffix)]
            suffix = '.xunit'
            if file_name.endswith(suffix):
                file_name = file_name[0:-len(suffix)]
        testname = '%s.%s' % (folder_name, file_name)

        xml = get_xunit_content(report, testname, time.time() - start_time)
        path = os.path.dirname(os.path.abspath(args.xunit_file))
        if not os.path.exists(path):
            os.makedirs(path)
        with open(args.xunit_file, 'w', encoding='utf-8') as f:
            f.write(xml)

    return rc


def add_missing_header(file_descriptors, name, license_, verbose):
    copyright_ = 'Copyright %d %s' % (int(time.strftime('%Y')) - 1 + 1, name)
    header = license_.file_headers[0].format(**{
        'copyright': copyright_,
        'copyright_holder': name})
    lines = header.splitlines()

    if verbose:
        print('Adding the following copyright / license header or repository level files:')
        for line in lines:
            print('+', line)
        print()

    for path in sorted(file_descriptors.keys()):
        file_descriptor = file_descriptors[path]
        skip = False

        if file_descriptor.filetype == SOURCE_FILETYPE:
            # ignore empty files
            if not file_descriptor.content:
                skip = True
            # ignore files which already have a header
            if file_descriptor.copyright_identifiers:
                skip = True
        elif file_descriptor.exists:
            # ignore non-source files if they already exist
            skip = True

        if skip:
            if verbose:
                print(' ', file_descriptor.path)
            continue

        if file_descriptor.filetype == SOURCE_FILETYPE:
            print('*', file_descriptor.path)
            add_header(file_descriptor, header)

        elif file_descriptor.filetype == CONTRIBUTING_FILETYPE:
            print('+', file_descriptor.path)
            with open(file_descriptor.path, 'w', encoding='utf-8') as h:
                h.write(license_.contributing_files[0])

        elif file_descriptor.filetype == LICENSE_FILETYPE:
            print('+', file_descriptor.path)
            with open(file_descriptor.path, 'w', encoding='utf-8') as h:
                h.write(license_.license_files[0])

        else:
            assert False, 'Unknown filetype: ' + file_descriptor.filetype


def add_copyright_year(file_descriptors, new_years, verbose):
    if verbose:
        print(f'Adding {",".join(map(str, new_years))} to existing copyright notices:')
        print()

    for path in sorted(file_descriptors.keys()):
        file_descriptor = file_descriptors[path]

        # ignore files which do not have a header
        if not getattr(file_descriptor, 'copyright_identifiers', None):
            continue

        index = scan_past_coding_and_shebang_lines(file_descriptor.content)
        index = scan_past_empty_lines(file_descriptor.content, index)

        if file_descriptor.filetype == SOURCE_FILETYPE:
            block, block_offset = get_comment_block(file_descriptor.content, index)
            if not block:
                assert False, "Could not find comment block in file '%s'" % file_descriptor.path
        else:
            block = file_descriptor.content[index:]
            block_offset = 0
        copyrights, years_spans, _, _ = search_copyright_information(block)
        if copyrights is None:
            assert False, "Could not find copyright information in file '%s'" % \
                file_descriptor.path

        for years_span in years_spans:
            # skip if all new years are already included
            years = get_years_from_string(block[years_span[0]:years_span[1]])
            if all((new_year in years) for new_year in new_years):
                if verbose:
                    print(' ', file_descriptor.path)
                continue
            print('*' if file_descriptor.exists else '+', file_descriptor.path)

            for new_year in new_years:
                years.add(new_year)
            years_string = get_string_from_years(years)

            # overwrite previous years with new years
            offset = index + block_offset
            global_years_span = [offset + years_span[0], offset + years_span[1]]
            content = file_descriptor.content[:global_years_span[0]] + years_string + \
                file_descriptor.content[global_years_span[1]:]

            with open(file_descriptor.path, 'w', encoding='utf-8') as h:
                h.write(content)


def get_years_from_string(content):
    # remove all whitespaces
    content = re.sub(r'\s', '', content)
    # split items by comma
    items = content.split(',')

    years = set()
    for item in items:
        # each item can be a plain year or a range of years
        parts = item.split('-', 1)
        if len(parts) == 1:
            # plain year
            years.add(int(parts[0]))
        else:
            # year range
            start_year = int(parts[0])
            end_year = int(parts[1])
            assert end_year >= start_year
            for year in range(start_year, end_year + 1):
                years.add(year)
    return years


def get_string_from_years(years):
    ranges = []
    for _, iterable in groupby(enumerate(sorted(years)), lambda x: x[1] - x[0]):
        r = list(iterable)
        if len(r) == 1:
            ranges.append(str(r[0][1]))
        else:
            ranges.append('%d-%d' % (r[0][1], r[-1][1]))
    return ', '.join(ranges)


def add_header(file_descriptor, header):
    """
    Add the copyright / license message to a file.

    The copyright / license is placed below an optional shebang line as
    well as an optional coding line.
    It is preceded by an empty line if not at the beginning of the file
    and always followed by an empty line.
    """
    begin_index = scan_past_coding_and_shebang_lines(file_descriptor.content)
    end_index = scan_past_empty_lines(file_descriptor.content, begin_index)

    # inject copyright message
    comment = get_comment(file_descriptor.path, header)
    inserted_block = '%s\n\n' % comment
    if begin_index > 0:
        inserted_block = '\n' + inserted_block
    content = file_descriptor.content[:begin_index] + inserted_block + \
        file_descriptor.content[end_index:]

    # output beginning of file for debugging
    # index = end_index + len(inserted_block)
    # for _ in range(3):
    #     index = get_index_of_next_line(content, index)
    # print('<<<')
    # print(content[:index - 1])
    # print('>>>')

    with open(file_descriptor.path, 'w', encoding='utf-8') as h:
        h.write(content)


def get_comment(filename, msg):
    if filename.endswith('.cmake') or filename.endswith('.py'):
        line_prefix = '#'
    else:
        line_prefix = '//'

    comment = ''
    index = 0
    while True:
        new_index = get_index_of_next_line(msg, index)
        if new_index == index:
            break
        comment += line_prefix
        line = msg[index:new_index]
        if line.splitlines()[0]:
            comment += ' '
        comment += line
        index = new_index
    return comment


def get_xunit_content(report, testname, elapsed):
    test_count = len(report)
    error_count = len([r for r in report if not r[1]])
    data = {
        'testname': testname,
        'test_count': test_count,
        'error_count': error_count,
        'time': '%.3f' % round(elapsed, 3),
    }
    xml = """<?xml version="1.0" encoding="UTF-8"?>
<testsuite
  name="%(testname)s"
  tests="%(test_count)d"
  errors="0"
  failures="%(error_count)d"
  time="%(time)s"
>
""" % data

    for (filename, no_error, message) in report:

        data = {
            'quoted_filename': quoteattr(filename),
            'testname': testname,
            'escaped_message': escape(message),
        }
        if not no_error:
            # report missing / unknown copyright / license as a failing testcase
            xml += """  <testcase
    name=%(quoted_filename)s
    classname="%(testname)s"
  >
      <failure message="%(escaped_message)s"/>
  </testcase>
""" % data

        else:
            # if there is a known copyright / license report a single successful test
            xml += """  <testcase
    name=%(quoted_filename)s
    classname="%(testname)s"/>
""" % data

    # output list of checked files
    data = {
        'escaped_files': escape(''.join(['\n* %s' % r[0] for r in report])),
    }
    xml += """  <system-out>Checked files:%(escaped_files)s</system-out>
""" % data

    xml += '</testsuite>\n'
    return xml


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2015 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os

from ament_copyright import ALL_FILETYPES
from ament_copyright import SOURCE_FILETYPE


def get_files(paths, extensions, exclude_patterns, skip_package_level_setup_py=True):
    excludes = []
    for exclude_pattern in exclude_patterns:
        excludes.extend(glob.glob(exclude_pattern))
    excludes = {os.path.realpath(x) for x in excludes}

    files = {}
    for path in paths:
        if os.path.isdir(path):
            if is_repository_root(path):
                add_files_for_all_filetypes(path, files)
            for dirpath, dirnames, filenames in os.walk(path):
                if 'AMENT_IGNORE' in dirnames + filenames:
                    dirnames[:] = []
                    continue
                if is_repository_root(dirpath):
                    add_files_for_all_filetypes(dirpath, files)

                # ignore folder starting with . or _
                dirnames[:] = [d for d in dirnames if d[0] not in ['.', '_']]
                dirnames.sort()

                # select files by extension
                for filename in sorted(filenames):
                    # skip package level setup.py file
                    if (
                        skip_package_level_setup_py and
                        filename == 'setup.py' and
                        os.path.exists(os.path.join(dirpath, 'package.xml'))
                    ):
                        continue
                    if match_filename(filename, extensions):
                        filepath = os.path.join(dirpath, filename)
                        if os.path.realpath(filepath) not in excludes:
                            files[os.path.join(dirpath, filename)] = SOURCE_FILETYPE

        if os.path.isfile(path) and match_filename(path, extensions):
            if os.path.realpath(path) not in excludes:
                files[path] = SOURCE_FILETYPE

        if is_repository_root(os.path.dirname(path)):
            basename = os.path.basename(path)
            for filetype, filename in ALL_FILETYPES.items():
                if filename == basename:
                    files[path] = filetype
                    break
    return {os.path.normpath(path): filetype for path, filetype in files.items()}


def is_repository_root(path):
    """Check if the path is the root of a git or mercurial repository."""
    return (
        os.path.exists(os.path.join(path, '.git')) or
        os.path.exists(os.path.join(path, '.hg'))
    )


def match_filename(filename, extensions):
    """Check if the filename has one of the extensions."""
    _, ext = os.path.splitext(filename)
    return ext in ('.%s' % e for e in extensions)


def add_files_for_all_filetypes(path, files):
    for filetype, filename in ALL_FILETYPES.items():
        if filename is None:
            continue
        files[os.path.join(path, filename)] = filetype



# Copyright 2015 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata


COPYRIGHT_GROUP = 'ament_copyright.copyright_name'
LICENSE_GROUP = 'ament_copyright.license'

SOURCE_FILETYPE = 1
CONTRIBUTING_FILENAME = 'CONTRIBUTING.md'
CONTRIBUTING_FILETYPE = 2
LICENSE_FILENAME = 'LICENSE'
LICENSE_FILETYPE = 3

ALL_FILETYPES = {
    SOURCE_FILETYPE: None,
    CONTRIBUTING_FILETYPE: CONTRIBUTING_FILENAME,
    LICENSE_FILETYPE: LICENSE_FILENAME,
}

UNKNOWN_IDENTIFIER = '<unknown>'


def get_copyright_names():
    names = {}
    entry_points = importlib_metadata.entry_points()
    if hasattr(entry_points, 'select'):
        copyright_groups = entry_points.select(group=COPYRIGHT_GROUP)
    else:
        copyright_groups = entry_points.get(COPYRIGHT_GROUP, [])
    for entry_point in copyright_groups:
        assert entry_point.name != UNKNOWN_IDENTIFIER, \
            "Invalid entry point name '%s'" % entry_point.name
        name = entry_point.load()
        names[entry_point.name] = name
    return names


def get_licenses():
    licenses = {}
    entry_points = importlib_metadata.entry_points()
    if hasattr(entry_points, 'select'):
        license_groups = entry_points.select(group=LICENSE_GROUP)
    else:
        license_groups = entry_points.get(LICENSE_GROUP, [])
    for entry_point in license_groups:
        assert entry_point.name != UNKNOWN_IDENTIFIER, \
            "Invalid entry point name '%s'" % entry_point.name
        licenses[entry_point.name] = entry_point.load()
    return licenses



# Copyright 2015 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# to add more licenses do not extend this file
# instead create a separate package and register custom licenses as entry points

from collections import namedtuple
import os

LicenseEntryPoint = namedtuple(
    'LicenseEntryPoint', ['name', 'spdx', 'file_headers', 'license_files', 'contributing_files'])

TEMPLATE_DIRECTORY = os.path.join(os.path.dirname(__file__), 'template')


def read_data(path, name, prefix, license_type):
    path_template = os.path.join(path, prefix + '_' + license_type + '_%d.txt')
    data = []

    with open(os.path.join(path, prefix + '_' + license_type + '.txt'), 'r') as h:
        data.append(h.read())

    index = 0
    while True:
        try:
            with open(path_template % index, 'r') as h:
                data.append(h.read())
                index += 1
        except OSError:
            break

    return data


def read_license_data(path, name, spdx, prefix):
    file_headers = read_data(path, name, prefix, 'header')
    license_files = read_data(path, name, prefix, 'license')
    contributing_files = read_data(path, name, prefix, 'contributing')

    return LicenseEntryPoint(name, spdx, file_headers, license_files, contributing_files)


# The SPDX identifier (the 3rd argument) comes from the official list at https://spdx.org/licenses/

apache2 = read_license_data(TEMPLATE_DIRECTORY,
                            'Apache License, Version 2.0',
                            'Apache-2.0',
                            'apache2')
boost1 = read_license_data(TEMPLATE_DIRECTORY,
                           'Boost Software License - Version 1.0',
                           'BSL-1.0',
                           'boost1')
bsd2 = read_license_data(TEMPLATE_DIRECTORY,
                         'BSD License 2.0',
                         'BSD-2.0',
                         'bsd2')
bsd_3clause = read_license_data(TEMPLATE_DIRECTORY,
                                '3-Clause BSD License',
                                'BSD-3-Clause',
                                'bsd_3clause')
bsd_2clause = read_license_data(TEMPLATE_DIRECTORY,
                                '2-Clause BSD License',
                                'BSD-2-Clause',
                                'bsd_2clause')
mit = read_license_data(TEMPLATE_DIRECTORY,
                        'MIT License',
                        'MIT',
                        'mit')
mit0 = read_license_data(TEMPLATE_DIRECTORY,
                         'MIT-0 License',
                         'MIT-0',
                         'mit0')
gplv3 = read_license_data(TEMPLATE_DIRECTORY,
                          'GNU General Public License 3.0',
                          'GPL-3.0-only',
                          'gplv3')
lgplv3 = read_license_data(TEMPLATE_DIRECTORY,
                           'GNU Lesser General Public License 3.0',
                           'LGPL-3.0-only',
                           'lgplv3')



# Copyright 2015 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from ament_copyright import ALL_FILETYPES
from ament_copyright import CONTRIBUTING_FILETYPE
from ament_copyright import get_copyright_names
from ament_copyright import get_licenses
from ament_copyright import LICENSE_FILETYPE
from ament_copyright import SOURCE_FILETYPE
from ament_copyright import UNKNOWN_IDENTIFIER


class CopyrightDescriptor:

    def __init__(self, name, year_range):
        self.name = name
        self.year_range = year_range

    def __str__(self):
        s = self.name
        if self.year_range:
            s += ' (%s)' % self.year_range
        return s


class FileDescriptor:

    def __init__(self, filetype, path):
        self.filetype = filetype
        self.path = path
        self.exists = os.path.exists(path)
        self.content = None
        self.license_identifier = UNKNOWN_IDENTIFIER

    def read(self):
        if not self.exists:
            return
        with open(self.path, 'r', encoding='utf-8') as h:
            self.content = h.read()

    def parse(self):
        raise NotImplementedError()

    def identify_license(self, content, license_part, licenses=None):
        if content is None:
            return
        if licenses is None:
            licenses = get_licenses()
        formatted_content = remove_formatting(content)

        for name, license_ in licenses.items():
            templates = getattr(license_, license_part)
            for template in templates:
                last_index = -1
                formatted_template = remove_formatting(template)
                template_sections = split_template(formatted_template,
                                                   ['{copyright_holder}', '{copyright}'])
                for license_section in template_sections:
                    # OK, now look for each section of the license in the incoming
                    # content.
                    index = formatted_content.find(license_section.strip())
                    if index == -1 or index <= last_index:
                        # Some part of the license is not in the content, or the license
                        # is rearranged, this license doesn't match.
                        break
                    last_index = index
                else:
                    # We found the license, so set it
                    self.license_identifier = name
                    break


class SourceDescriptor(FileDescriptor):

    def __init__(self, path):
        super(SourceDescriptor, self).__init__(SOURCE_FILETYPE, path)

        self.copyrights = []

        self.copyright_identifiers = []

    def identify_copyright(self):
        known_copyrights = get_copyright_names()
        for c in self.copyrights:
            found_name = c.name
            for identifier, name in known_copyrights.items():
                if name == found_name:
                    self.copyright_identifiers.append(identifier)
                    break
            else:
                self.copyright_identifiers.append(UNKNOWN_IDENTIFIER)

    def parse(self):
        self.read()
        if not self.content:
            return

        # skip over coding and shebang lines
        index = scan_past_coding_and_shebang_lines(self.content)
        index = scan_past_empty_lines(self.content, index)

        # get first comment block without leading comment tokens
        block, _ = get_comment_block(self.content, index)
        copyrights, _, _, remaining_block = search_copyright_information(block)

        if len(copyrights) == 0:
            block = get_multiline_comment_block(self.content, index)
            copyrights, _, _, remaining_block = search_copyright_information(block)

        if len(copyrights) == 0:
            return

        self.copyrights = copyrights

        self.identify_copyright()

        content = '{copyright}' + remaining_block
        self.identify_license(content, 'file_headers')


class ContributingDescriptor(FileDescriptor):

    def __init__(self, path):
        super(ContributingDescriptor, self).__init__(CONTRIBUTING_FILETYPE, path)

    def parse(self):
        self.read()
        if not self.content:
            return

        self.identify_license(self.content, 'contributing_files')


class LicenseDescriptor(FileDescriptor):

    def __init__(self, path):
        super(LicenseDescriptor, self).__init__(LICENSE_FILETYPE, path)

    def parse(self):
        self.read()
        if not self.content:
            return

        self.identify_license(self.content, 'license_files')


def parse_file(path):
    filetype = determine_filetype(path)
    if filetype == SOURCE_FILETYPE:
        d = SourceDescriptor(path)
    elif filetype == CONTRIBUTING_FILETYPE:
        d = ContributingDescriptor(path)
    elif filetype == LICENSE_FILETYPE:
        d = LicenseDescriptor(path)
    else:
        return None
    d.parse()
    return d


def determine_filetype(path):
    basename = os.path.basename(path)
    for filetype, filename in ALL_FILETYPES.items():
        if basename == filename:
            return filetype
    return SOURCE_FILETYPE


def search_copyright_information(content):
    if content is None:
        return [], content
    # regex for matching years or year ranges (yyyy-yyyy) separated by colons
    year = r'\d{4}'
    year_range = '%s-%s' % (year, year)
    year_or_year_range = '(?:%s|%s)' % (year, year_range)
    pattern = r'^[^\n\r]?\s*(?:\\copyright\s*)?' \
              r'copyright(?:\s+\(c\))?\s+(%s(?:,\s*%s)*),?\s+([^\n\r]+)$' % \
        (year_or_year_range, year_or_year_range)
    regex = re.compile(pattern, re.DOTALL | re.MULTILINE | re.IGNORECASE)

    copyrights = []
    years_spans = []
    name_spans = []
    while True:
        match = regex.search(content)
        if not match:
            break
        years_span, name_span = match.span(1), match.span(2)
        years_spans.append(years_span)
        name_spans.append(name_span)
        years = content[years_span[0]:years_span[1]]
        name = content[name_span[0]:name_span[1]]
        copyrights.append(CopyrightDescriptor(name, years))
        content = content[name_span[1]:]

    return copyrights, years_spans, name_spans, content


def scan_past_coding_and_shebang_lines(content):
    index = 0
    while (
        is_comment_line(content, index) and
        (is_coding_line(content, index) or
         is_shebang_line(content, index))
    ):
        index = get_index_of_next_line(content, index)
    return index


def get_index_of_next_line(content, index):
    index_n = content.find('\n', index)
    index_r = content.find('\r', index)
    index_rn = content.find('\r\n', index)
    indices = set()
    if index_n != -1:
        indices.add(index_n)
    if index_r != -1:
        indices.add(index_r)
    if index_rn != -1:
        indices.add(index_rn)
    if not indices:
        return len(content)
    index = min(indices)
    if index == index_rn:
        return index + 2
    return index + 1


def is_comment_line(content, index):
    # skip over optional BOM
    if index == 0 and content[0] == '\ufeff':
        index = 1
    return content[index] == '#' or content[index:index + 1] == '//'


def is_coding_line(content, index):
    end_index = get_index_of_next_line(content, index)
    line = content[index:end_index]
    return 'coding=' in line or 'coding:' in line


def is_shebang_line(content, index):
    # skip over optional BOM
    if index == 0 and content[0] == '\ufeff':
        index = 1
    return content[index:index + 2] == '#!'


def get_comment_block(content, index):
    # regex for matching the beginning of the first comment
    # check for doxygen comments (///) before regular comments (//)
    pattern = '^(#|///|//)'
    # also accept BOM if present
    if index == 0 and content[0] == '\ufeff':
        pattern = pattern[0] + '\ufeff' + pattern[1:]
    regex = re.compile(pattern, re.MULTILINE)

    match = regex.search(content, index)
    if not match:
        return None, None
    comment_token = match.group(1)
    start_index = match.start(1)

    end_index = start_index
    while True:
        end_index = get_index_of_next_line(content, end_index)
        if content[end_index:end_index + len(comment_token)] != comment_token:
            break

    block = content[start_index:end_index]
    lines = block.splitlines()
    lines = [line[len(comment_token) + 1:] for line in lines]

    return '\n'.join(lines), start_index + len(comment_token) + 1


def get_multiline_comment_block(content, index):
    patterns = [('^(/[*])', '([*]/)$'),
                ('^(<!--)', '(-->)$')]
    for pattern_pair in patterns:
        start_pattern, end_pattern = pattern_pair
        # find the first match of the comment start token
        # also accept BOM if present
        if index == 0 and content[0] == '\ufeff':
            start_pattern = start_pattern[0] + '\ufeff' + start_pattern[1:]
        start_regex = re.compile(start_pattern, re.MULTILINE)
        start_match = start_regex.search(content, index)
        if not start_match:
            continue
        start_index = start_match.start(1)

        # find the first match of the comment end token
        end_regex = re.compile(end_pattern, re.MULTILINE)
        end_match = end_regex.search(content, index)
        if not end_match:
            continue
        end_index = end_match.start(1)

        # collect all lines between start and end (open interval) and strip out any common prefix
        block = content[start_index:end_index]
        block_lines = block.splitlines()
        if len(block_lines) == 1:
            prefixed_lines = block_lines
        elif len(block_lines) == 2:
            prefixed_lines = block_lines[1:]
        else:
            prefixed_lines = block_lines[1:-1]

        if len(prefixed_lines) > 1:
            line_prefix = os.path.commonprefix(prefixed_lines)
            lines = [line[len(line_prefix):] for line in prefixed_lines]
        else:
            # Single-line header does not have a common prefix to strip out
            lines = prefixed_lines

        return '\n'.join(lines)
    return None


def scan_past_empty_lines(content, index):
    while is_empty_line(content, index):
        index = get_index_of_next_line(content, index)
    return index


def is_empty_line(content, index):
    return get_index_of_next_line(content, index) == index + 1


def remove_formatting(text):
    return ' '.join(filter(None, text.split()))


# Flat list of sections split on all separators provided
def split_template(sections, separators):
    if not isinstance(sections, list):
        return split_template([sections], separators)
    elif len(separators) > 1:
        return sum([split_template([section], separators[0:1]) for section
                    in sum([split_template([section], separators[1:])
                            for section in sections], [])], [])
    else:
        return list(filter(lambda s: len(s) > 0,
                           sum([section.split(separators[0]) for section in sections], [])))
